@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2010-2016 Phusion Holding B.V.
+ *  Copyright (c) 2010-2018 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -733,6 +733,64 @@ connectToServer(NConnect_State &state) {
 	}
 }
 
+bool
+pingTcpServer(const StaticString &host, unsigned int port, unsigned long long *timeout) {
+	TRACE_POINT();
+	NTCP_State state;
+
+	setupNonBlockingTcpSocket(state, host, port, __FILE__, __LINE__);
+
+	try {
+		if (connectToTcpServer(state)) {
+			return true;
+		}
+	} catch (const SystemException &e) {
+		if (e.code() == ECONNREFUSED) {
+			return false;
+		} else {
+			throw e;
+		}
+	}
+
+	// Cannot connect to the port yet, but that may not mean the
+	// port is unavailable. So poll the socket.
+
+	bool connectable;
+	try {
+		connectable = waitUntilWritable(state.fd, timeout);
+	} catch (const SystemException &e) {
+		throw SystemException("Error polling TCP socket "
+			+ host + ":" + toString(port), e.code());
+	}
+	if (!connectable) {
+		// Timed out. Assume port is not available.
+		return false;
+	}
+
+	// Try to connect the socket one last time.
+
+	try {
+		return connectToTcpServer(state);
+	} catch (const SystemException &e) {
+		if (e.code() == ECONNREFUSED) {
+			return false;
+		} else if (e.code() == EISCONN || e.code() == EINVAL) {
+			#ifdef __FreeBSD__
+				// Work around bug in FreeBSD (discovered on
+				// January 20 2013 in daemon_controller)
+				return false;
+			#else
+				throw e;
+			#endif
+
+			// Never reached, shut up compiler warning.
+			return false;
+		} else {
+			throw e;
+		}
+	}
+}
+
 SocketPair
 createUnixSocketPair(const char *file, unsigned int line) {
 	int fds[2];
@@ -1296,32 +1354,23 @@ safelyClose(int fd, bool ignoreErrors) {
 	}
 }
 
-string
-readAll(const string &filename) {
-	FILE *f = fopen(filename.c_str(), "rb");
-	if (f != NULL) {
-		StdioGuard guard(f, NULL, 0);
-		return readAll(fileno(f));
-	} else {
-		int e = errno;
-		throw FileSystemException("Cannot open '" + filename + "' for reading",
-			e, filename);
-	}
-}
-
-string
-readAll(int fd) {
+pair<string, bool>
+readAll(int fd, size_t maxSize) {
 	string result;
 	char buf[1024 * 32];
 	ssize_t ret;
-	while (true) {
+	bool eofReached = false;
+
+	while (result.size() < maxSize) {
 		do {
 			ret = read(fd, buf, sizeof(buf));
 		} while (ret == -1 && errno == EINTR);
 		if (ret == 0) {
+			eofReached = true;
 			break;
 		} else if (ret == -1) {
 			if (errno == ECONNRESET) {
+				eofReached = true;
 				break;
 			} else {
 				int e = errno;
@@ -1331,7 +1380,8 @@ readAll(int fd) {
 			result.append(buf, ret);
 		}
 	}
-	return result;
+
+	return make_pair(result, eofReached);
 }
 
 
